// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* A3StlMod.cpp *#*#*#*#*#*#*#*#*#*#* (C) 2022 DekTec
//
//

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <stdio.h>
#include <sstream>
#include "DTAPI.h"
#include "A3StlMod.h"
#include "DtOpt.h"


#ifdef WINBUILD
    #include <Windows.h>
    #include <conio.h>
#else
    #include <stdlib.h>
    #include <string.h>
    #include <unistd.h>
    #include <limits.h>
    #include <termios.h>
#endif

#ifndef WINBUILD
void ChangeTerminalMode(int dir)
{
    static struct termios  OldT, NewT;

    if (dir == 1)
    {
        tcgetattr(STDIN_FILENO, &OldT);
        NewT = OldT;
        NewT.c_lflag &= ~(ICANON | ECHO);
        tcsetattr(STDIN_FILENO, TCSANOW, &NewT);
    } else {
        tcsetattr(STDIN_FILENO, TCSANOW, &OldT);
    }
}
int  _kbhit()
{
    struct timeval  tv;
    fd_set  rdfs;

    tv.tv_sec = 0;
    tv.tv_usec = 0;

    FD_ZERO(&rdfs);
    FD_SET(STDIN_FILENO, &rdfs);

    select(STDIN_FILENO+1, &rdfs, NULL, NULL, &tv);
    return FD_ISSET(STDIN_FILENO, &rdfs);
}
#endif


// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- A3StlMod Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#define A3STLMOD_VERSION_MAJOR        1
#define A3STLMOD_VERSION_MINOR        0
#define A3STLMOD_VERSION_BUGFIX       3


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Error messages -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// Command line errors
const char c_CleRfFreqNotDefined[]      = "Frequency not specified";
const char c_CleInvalidBandwidth[]      = "Invalid bandwidth specified";
const char c_CleIpAddressNotDefined[]   = "IP-address not specified";
const char c_CleIpAddressInvalid[]      = "Invalid IP multicast/unicast address";

// General errors
const char c_ErrDtapiDvcScanFailed[]    = "DtapiDeviceScan failed (ERROR: %s)";
const char c_ErrNoOutputFound[]         = "No output device in the system";
const char c_ErrNoSuchPort[]            = "The %s has no port %d";
const char c_ErrHwScanFailed[]          = "HwFuncScan failed (ERROR: %s)";
const char c_ErrNoPort[]                = "The %s port %d it not an suitable port";
const char c_ErrNoSuitablePort[]        = "No suitable port found on the %s";
const char c_ErrFailToAttachToChan[]    = "Can't attach to the channel (ERROR: %s)";
const char c_ErrFailToAttachToType[]    = "Failed to attach to the %s (ERROR: %s)";
const char c_ErrDvcNumWithoutType[]     = "Specifying a device number without a type is not supported";
const char c_ErrFailGetFlags[]          = "GetFlags failed (ERROR: %s)";
const char c_ErrFailSetTxControl[]      = "SetTxControl failed (ERROR: %s)";
const char c_ErrFailSetModControl[]     = "SetModControl failed (ERROR: %s)";
const char c_ErrFailSetRfControl[]      = "SetRfControl failed (ERROR: %s)";
const char c_ErrFailSetIpPars[]         = "Setting IP-parameters failed (function:%s returned %d)";
const char c_ErrFailIpBind[]            = "Network interface binding to %s failed (ERROR: %d)";
const char c_ErrFailAddMca[]            = "Adding to multicast group %s failed (ERROR: %d)";
const char c_ErrFailAddMcaSrc[]         = "Adding to multicast group %s and source %s failed (ERROR: %d)";
const char c_ErrFailWrite[]             = "Write failed (ERROR: %s)";
const char c_ErrFailSetOutputLevel[]    = "Failed to set output level (ERROR: %s)";
const char c_ErrFailGetFifoLoad[]       = "Failed to get Fifo load (ERROR: %s)";
const char c_ErrFailSetFifoSize[]       = "Failed to set Fifo size (ERROR: %s)";
const char c_ErrCpuUnderflow[]          = "CPU underflow detected";
const char c_ErrDmaUnderflow[]          = "Dma underflow detected";
const char c_ErrFifoUnderflow[]         = "Fifo underflow detected";

static DtOptItem*  g_CmdOptions;
static int  g_NumCmdOptions;

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- wtoi -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
static int wtoi(wstring Str)
{
    wistringstream Stream(Str);
    int  Result;
    Stream >> Result;
    return Result;
}


//+=+=+=+=+=+=+=+=+=+=+=+=+=+ CommandLineParams implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.- CommandLineParams::ParseCommandLine -.-.-.-.-.-.-.-.-.-.-.-.-.
//
void CommandLineParams::ParseCommandLine(int argc, char* argv[])
{
    static DtOptItem CmdOptions[] = {
        DtOptItem(L"t",   m_DvcType, -1, L"Device type to use (315, 2115 or 2116)", 315, 2116),
        DtOptItem(L"n",   m_DvcNum, 1, L"Device number to use (default: 1)", 1, 99),
        DtOptItem(L"i",   m_Port, -1, L"Port number of the output channel to use", 1, 8),
        DtOptItem(L"mf",  m_CarrierFreq, 0.0, L"Carrier frequency in MHz", 20.0, 2300.0),
        DtOptItem(L"mb",  m_Bandwidth, 6, L"Output bandwidth in MHz (default: 6 MHz)", 6, 8),
        DtOptItem(L"ml",  m_OutpLevel, -27.5, L"Output level in dBm (default: -27.5dBm)", -35.0, 0.0),
        DtOptItem(L"ipa", m_IpAddr, L"IP (multicast) address/port (e.g. 0.0.0.0:1234 or 239.2.2.2:5000)"),
        DtOptItem(L"ipitf", m_IpItf, L"Network interface to use (e.g. 192.168.2.1)"),
        DtOptItem(L"ipsrc", m_IpSrc, L"Specific IP source address (e.g. 10.5.120.17)"),
        DtOptItem(L"a",   m_ActiveInd, false, L"Show the activity indicator"),
        DtOptItem(L"?",   m_ShowHelp, false, L"Display this help"),
    };

    g_CmdOptions = CmdOptions;
    g_NumCmdOptions = sizeof(CmdOptions) / sizeof(CmdOptions[0]);

    list<wstring>  FreeArgs;
    DtOptItem::ParseOpt(CmdOptions, argc, argv, FreeArgs);

    // Show help?
    if (m_ShowHelp.IsSet())
        return;

    // Carrier frequency must be specified    
    if (!m_CarrierFreq.IsSet())
        throw Exc(c_CleRfFreqNotDefined);
    
    // IP-address must be specified
    if (!m_IpAddr.IsSet())
        throw Exc(c_CleIpAddressNotDefined);

    memset(&m_IpPars, 0, sizeof(m_IpPars));

    // Parse IP-address
    wstring  IpAddr = m_IpAddr.ToString();
    size_t  Pos = IpAddr.find(L":");
    if (Pos == wstring::npos)
        throw Exc(c_CleIpAddressInvalid);

    int  IpPort = wtoi(IpAddr.c_str() + Pos + 1);
    if (IpPort<0 || IpPort>0xFFFF)
        throw Exc(c_CleIpAddressInvalid, "ipa");
    IpAddr.resize(Pos);

    // Parse optional IP source address
    const wchar_t* pSrcAddr = nullptr;
    wstring  SrcAddr;
    if (m_IpSrc.IsSet())
    {
        SrcAddr = m_IpSrc.ToString();
        pSrcAddr = SrcAddr.c_str();
    }
    DTAPI_RESULT dr = ::DtapiInitDtIpParsFromIpString(m_IpPars, IpAddr.c_str(), pSrcAddr);
    m_IpPars.m_Port = IpPort;
    m_IpPars.m_Protocol = DTAPI_PROTO_UDP;

    // Get optional network interface address
    // Default any interface
    m_Nic[0] = m_Nic[1] = m_Nic[2] = m_Nic[3] = 0;
    if (m_IpItf.IsSet())
    {
        // Parse IP-interface address
        wstring ItfAddr = m_IpItf.ToString();
        int Flags = 0;
        dr = ::DtapiIpAddr2ByteArray(ItfAddr.c_str(), m_Nic, Flags);
        if (dr != DTAPI_OK)
            throw Exc(c_CleIpAddressInvalid, "ipitf");
    }
}


// =+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Log functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=
//

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Log -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void Log(const char* pMessage)
{
    // Under linux we can't mix printf and wprintf, so convert
    // to wstring and always use wprintf
    string  Str(pMessage);
    Log(wstring(Str.begin(), Str.end()).c_str());
}
void Log(const wchar_t* pMessage)
{
    // Print message and add new line
    wprintf(pMessage); wprintf(L"\n");
}
// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- LogF -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Overload with optional parameters
//
void LogF(const char* pMessage, ... )
{
    char sz[512];

    va_list  ArgList;
    va_start(ArgList, pMessage);
    _vsnprintf(sz, sizeof(sz)-1, pMessage, ArgList);
    va_end(ArgList);
    Log(sz);
}
// Overload with optional parameters
void LogF(const wchar_t* pMessage, ... )
{
    wchar_t sz[512];

    va_list  ArgList;
    va_start(ArgList, pMessage);
    _vsnwprintf(sz, (sizeof(sz)/sizeof(wchar_t))-1, pMessage, ArgList);
    va_end(ArgList);
    Log(sz);
}


//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ class NwUtils +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- NwUtils::IsMulticastAddressV4 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
bool  NwUtils::IsMulticastAddressV4(const unsigned char* pIpAddr)
{
    return pIpAddr[0]>=224 && pIpAddr[0]<=239;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- NwUtils::IsIpAddressEmpty -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool NwUtils::IsIpAddressEmpty(const unsigned char* pIpAddress)
{
    for (int i=0; i<4; i++)
        if (pIpAddress[i] != 0)
            return false;
    return true;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- NwUtils::Ip2InAddr -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  NwUtils::Ip2InAddr(const unsigned char* pIpAddress, in_addr* pInAddr)
{
    unsigned char*  pAddr = (unsigned char*)pInAddr;
    for (int i=0; i<4; i++) 
        pAddr[i] = pIpAddress[i];
}


//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- NwUtils::InAddr2Ip -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  NwUtils::InAddr2Ip(const in_addr* pInAddr, unsigned char* IpAddress)
{
    unsigned char*  pAddr = (unsigned char*)pInAddr;
    for (int i=0; i<4; i++) 
        IpAddress[i] = pAddr[i];
}


// =+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ STLTP to ATSC 3.0 RF +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- StlModHelper::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void StlModHelper::Init(DtIpPars IpPars, uint8_t* pNwItf, DtOutpChannel* pA3Outp,
                                                 __int64 RfFreq, int Bandwidth, int Level)
{
    RxStltpInit(IpPars, pNwItf);
    TxRfInit(pA3Outp, RfFreq, Bandwidth, Level);
}


// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- StlModHelper::StlMod -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// STLTP -> ATSC 3.0 loop
//
void StlModHelper::StlMod(bool ActiveInd)
{
    if (m_pA3Outp==nullptr)
        throw Exc(c_ErrNoSuchPort);

    const char* StrBusy="|/-\\";
    int StrIdx = 0;
    int SendCount = 1;
    if (ActiveInd)
        printf("%c\r",StrBusy[(StrIdx++) & 0x3]);

    // Start reception
    const int MaxUdpSize = 64*1024;
    vector<char> Buffer(MaxUdpSize);
    DTAPI_RESULT dr;
    bool InSend = false;
    while (!_kbhit())
    {
        // Wait for next UDP packet
        int PacketLength = ::recv(m_RxSocket, Buffer.data(), MaxUdpSize, 0);
        if (PacketLength == SOCKET_ERROR)
        {
            DWORD Error = ::GetLastError();
            // Sleep 1ms to prevent an endless loop with reception failures all the time
            Sleep(1);
            continue;
        }

        // Update activity indicator
        if (ActiveInd && (SendCount++)%64==0)
            printf("%c\r", StrBusy[(StrIdx++) & 0x3]);

        int  StatusFlags=0, Latched=0;
        // Check current FIFO-load
        int FifoLoad;
        dr = m_pA3Outp->GetFifoLoad(FifoLoad);
        if (dr != DTAPI_OK)
            throw Exc(c_ErrFailGetFifoLoad, ::DtapiResult2Str(dr));
        if (FifoLoad+PacketLength < m_FifoSize)
        {
            // Forward to ATSC 3.0
            dr = m_pA3Outp->Write(Buffer.data(), PacketLength);
            if (dr != DTAPI_OK)
                throw Exc(c_ErrFailWrite, ::DtapiResult2Str(dr));
            FifoLoad += PacketLength;

            // If sufficient FIFO-load is built-up, go to SEND
            if (!InSend)
            {
                if (FifoLoad > 3*m_FifoSize/4)
                {
                    dr = m_pA3Outp->SetTxControl(DTAPI_TXCTRL_SEND);
                    if (dr != DTAPI_OK)
                        throw Exc(c_ErrFailSetTxControl, ::DtapiResult2Str(dr));
                    InSend = true;
                }
            }
            // Get status flags
            dr = m_pA3Outp->GetFlags(StatusFlags, Latched);
            if (dr != DTAPI_OK)
                throw Exc(c_ErrFailGetFlags, ::DtapiResult2Str(dr));
        } else
            Latched = DTAPI_TX_FIFO_OVF;

        if (Latched != 0)
        {
            // Log error
            if ((Latched & DTAPI_TX_FIFO_OVF) != 0)
                LogF(L"Overflow; STL rate too high.");
            else if ((Latched & DTAPI_TX_FIFO_UFL) != 0)
                LogF(L"Underflow; STL rate too low.");
            else if ((Latched & DTAPI_TX_DMA_UFL) != 0)
                LogF(L"Underflow; data transfer to the card too low");
            else if ((Latched & DTAPI_TX_CPU_UFL) != 0)
                LogF(L"Underflow; CPU performance too low");

            // Try to recover from this error
            dr = m_pA3Outp->SetTxControl(DTAPI_TXCTRL_IDLE);
            if (dr != DTAPI_OK)
                throw Exc(c_ErrFailSetTxControl, ::DtapiResult2Str(dr));
            dr = m_pA3Outp->SetTxControl(DTAPI_TXCTRL_HOLD);
            if (dr != DTAPI_OK)
                throw Exc(c_ErrFailSetTxControl, ::DtapiResult2Str(dr));
            InSend = false;
        }
    }
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- StlModHelper::Uninit -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void StlModHelper::Uninit()
{
    if (m_pA3Outp != nullptr)
        m_pA3Outp->SetTxControl(DTAPI_TXCTRL_IDLE);
    m_pA3Outp = nullptr;
    ::closesocket(m_RxSocket);
    WSACleanup();
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- StlModHelper::RxStltpInit -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Initializes the IP-receiver. Windows specific implementation.
// 
void StlModHelper::RxStltpInit(DtIpPars IpPars, uint8_t* pNwItf)
{   
    // Init socket library
    try
    {
        WSADATA WsaData;
        int RetVal;
        if (RetVal = ::WSAStartup( MAKEWORD(2,2), &WsaData) != 0 )
            throw Exc(c_ErrFailSetIpPars, "WSAStartup()", RetVal);

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Create the new socket -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
        m_RxSocket = ::socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if (m_RxSocket == SOCKET_ERROR)
            throw Exc(c_ErrFailSetIpPars, "socket()", WSAGetLastError());

        // Set receive buffer size (1M)
        int IntOption = 1024*1024;
        RetVal = ::setsockopt(m_RxSocket, SOL_SOCKET, SO_RCVBUF, (char*)&IntOption,
                                                                       sizeof(IntOption));
        if (RetVal == SOCKET_ERROR) 
            throw Exc(c_ErrFailSetIpPars, "SetSockopt()", WSAGetLastError());

        // Set timeout (500ms)
        IntOption = 500;
        RetVal = ::setsockopt(m_RxSocket, SOL_SOCKET, SO_RCVTIMEO, (char*)&IntOption,
                                                                       sizeof(IntOption));
        if (RetVal == SOCKET_ERROR) 
            throw Exc(c_ErrFailSetIpPars, "SetSockopt()", WSAGetLastError());

        // Allow sharing of address
        BOOL BoolOption(TRUE);
        RetVal = ::setsockopt(m_RxSocket, SOL_SOCKET , SO_REUSEADDR, (char*)&BoolOption,
                                                                      sizeof(BoolOption));
        if (RetVal == SOCKET_ERROR) 
            throw Exc(c_ErrFailSetIpPars, "SetSockopt()", WSAGetLastError());


        // From: http://support.microsoft.com/default.aspx?scid=kb;en-us;263823
        // Disable "new behavior" using IOCTL: SIO_UDP_CONNRESET. Without this call,
        // recvfrom() can fail, repeatedly, after a bad sendto() call.
        DWORD dwBytesReturned = 0;
        BoolOption = FALSE;
        DWORD SIO_UDP_CONNRESET = -1744830452;
        RetVal =::WSAIoctl(m_RxSocket, SIO_UDP_CONNRESET, &BoolOption, sizeof(BoolOption),
                                                   NULL, 0, &dwBytesReturned, NULL, NULL);
        if (RetVal == SOCKET_ERROR) 
            throw Exc(c_ErrFailSetIpPars, "WSAIoctl()", WSAGetLastError());

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Connect the socket -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

        // Create a socket address
        SOCKADDR_IN Addr;
        ZeroMemory(&Addr, sizeof(Addr));
        Addr.sin_family = AF_INET;
        Addr.sin_addr.S_un.S_un_b.s_b1 = pNwItf[0];
        Addr.sin_addr.S_un.S_un_b.s_b2 = pNwItf[1];
        Addr.sin_addr.S_un.S_un_b.s_b3 = pNwItf[2];
        Addr.sin_addr.S_un.S_un_b.s_b4 = pNwItf[3];
        Addr.sin_port = htons(IpPars.m_Port);

        // Bind the socket
        RetVal = ::bind(m_RxSocket, (LPSOCKADDR)&Addr, sizeof(Addr));
        if (RetVal == SOCKET_ERROR)
        {
            if (NwUtils::IsIpAddressEmpty(pNwItf))
                throw Exc(c_ErrFailSetIpPars, "bind()", WSAGetLastError());
            else
            {
                char sz[256];
                DtapiIpAddr2Str(sz, sizeof(sz), pNwItf);
                throw Exc(c_ErrFailIpBind, sz,  WSAGetLastError());
            }
        }

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.- UNICAST/MCAST settings -.-.-.-.-.-.-.-.-.-.-.-.-.-.
        if (NwUtils::IsMulticastAddressV4(IpPars.m_Ip))
        {
            // MULTICAST
            if (NwUtils::IsIpAddressEmpty(IpPars.m_SrcFltIp))
            { 
                // Init FILTER and MCAST address
                ip_mreq  McastAddr;        // MCAST address
                ZeroMemory(&McastAddr, sizeof(McastAddr));
                // Set multicast group address
                NwUtils::Ip2InAddr(IpPars.m_Ip, &McastAddr.imr_multiaddr);
                // Set our own IP address
                NwUtils::Ip2InAddr(pNwItf, &McastAddr.imr_interface);
                // Add to MCAST group
                RetVal = ::setsockopt(m_RxSocket, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                                                    (char*)&McastAddr, sizeof(McastAddr));
                if (RetVal == SOCKET_ERROR)
                {
                    char sz[256];
                    DtapiIpAddr2Str(sz, sizeof(sz), IpPars.m_Ip);
                    throw Exc(c_ErrFailAddMca, sz, WSAGetLastError());
                }
            }
            else
            {
                // Add membership from only one source
                ip_mreq_source  McastReqSrc;
                ZeroMemory(&McastReqSrc, sizeof(McastReqSrc));

                NwUtils::Ip2InAddr(IpPars.m_Ip, &McastReqSrc.imr_multiaddr);
                NwUtils::Ip2InAddr(pNwItf, &McastReqSrc.imr_interface);
                NwUtils::Ip2InAddr(IpPars.m_SrcFltIp, &McastReqSrc.imr_sourceaddr);
                RetVal = ::setsockopt(m_RxSocket, IPPROTO_IP, IP_ADD_SOURCE_MEMBERSHIP, 
                                                (char*)&McastReqSrc, sizeof(McastReqSrc));
                if (RetVal == SOCKET_ERROR)
                {
                    char sz1[256];
                    DtapiIpAddr2Str(sz1, sizeof(sz1), IpPars.m_Ip);
                    char sz2[256];
                    DtapiIpAddr2Str(sz2, sizeof(sz2), IpPars.m_SrcFltIp);
                   throw Exc(c_ErrFailAddMcaSrc, sz1, sz2, WSAGetLastError());
                }
            }
        }
    }
    catch ( Exc e )
    {
        ::closesocket(m_RxSocket);
        WSACleanup();
        throw e;
    }
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- StlModHelper::TxRfInit -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void StlModHelper::TxRfInit(DtOutpChannel* pA3Outp, __int64 RfFreq, int Bandwidth,
                                                                                int Level)
{
    // For USB-device more buffering is required (= delay)
    m_FifoSize = 2*1024*1024;
    if (pA3Outp->Category() == DTAPI_CAT_USB)
        m_FifoSize = 16*1024*1024;

    // Set RF frequency
    DTAPI_RESULT dr = pA3Outp->SetRfControl(RfFreq);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrFailSetRfControl, ::DtapiResult2Str(dr));

    // Set RF level 
    dr = pA3Outp->SetOutputLevel(Level);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrFailSetOutputLevel, ::DtapiResult2Str(dr));

    // Start with transmitter in IDLE
    dr = pA3Outp->SetTxControl(DTAPI_TXCTRL_IDLE);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrFailSetTxControl, ::DtapiResult2Str(dr));

    // Initialize the modulator.
    DtAtsc3StltpPars StltpPars;
    switch (Bandwidth)
    {
    case 6: StltpPars.m_Bandwidth = DTAPI_ATSC3_6MHZ; break;
    case 7: StltpPars.m_Bandwidth = DTAPI_ATSC3_7MHZ; break;
    case 8: StltpPars.m_Bandwidth = DTAPI_ATSC3_8MHZ; break;
    default: throw Exc(c_CleInvalidBandwidth);
    }
    dr = pA3Outp->SetModControl(StltpPars);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrFailSetModControl, ::DtapiResult2Str(dr));

    // Keep delay low 
    dr = pA3Outp->SetFifoSize(m_FifoSize);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrFailSetFifoSize, ::DtapiResult2Str(dr));

    // Set transmitter to HOLD
    dr = pA3Outp->SetTxControl( DTAPI_TXCTRL_HOLD);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrFailSetTxControl, ::DtapiResult2Str(dr));

    m_pA3Outp = pA3Outp;
}


// +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ A3StlMod implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- A3StlMod::GetPortWithCaps -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  A3StlMod::GetPortWithCaps(DtDevice* pDvc, DtCaps& Caps, int& PortNr)
{
    PortNr = -1;
    DtHwFuncDesc HwFuncs[16];
    int  NumHwFuncs;
    DTAPI_RESULT  dr = pDvc->HwFuncScan(sizeof(HwFuncs)/sizeof(HwFuncs[0]),
                                                                     NumHwFuncs, HwFuncs);
    if (dr != DTAPI_OK)
        return false;
    for (int i=0; i<NumHwFuncs; i++)
    {
        if ((HwFuncs[i].m_Flags&Caps)==Caps)
        {
            PortNr = i+1;
            return true;
        }
    }
    return false;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Type2Name -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
static const char* Type2Name(int TypeNum)
{
    static char buf[32];
    if (TypeNum<200 || (TypeNum>=2000&&TypeNum<3000))
        sprintf(buf, "DTA-%d", TypeNum);
    else if (TypeNum>=200 && TypeNum <400)
        sprintf(buf, "DTU-%d", TypeNum);
    else if (TypeNum>3000)
        sprintf(buf, "DTE-%d", TypeNum);
    else
        sprintf(buf, "%d", TypeNum);
    return buf;
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- A3StlMod::AttachToPort -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void A3StlMod::AttachToPort(DtCaps& ReqCaps)
{
    DTAPI_RESULT dr;
    
    if (m_CmdLineParams.m_DvcType.IsSet())
    {
        dr = m_DtDvc.AttachToType(m_CmdLineParams.m_DvcType, m_CmdLineParams.m_DvcNum-1);
        if (dr != DTAPI_OK)
            throw Exc(c_ErrFailToAttachToType, Type2Name(m_CmdLineParams.m_DvcType),
                                                                     DtapiResult2Str(dr));
    } else {
        if (m_CmdLineParams.m_DvcNum.IsSet())
            throw Exc(c_ErrDvcNumWithoutType);
        DtDeviceDesc  DvcDescs[10];
        int  NumDvcs;
        dr = DtapiDeviceScan(sizeof(DvcDescs)/sizeof(DvcDescs[0]), NumDvcs, DvcDescs);
        if (dr != DTAPI_OK)
            throw Exc(c_ErrDtapiDvcScanFailed, DtapiResult2Str(dr));
        bool  FoundDvc = false;
        for (int i=0; i<NumDvcs; i++)
        {
            dr = m_DtDvc.AttachToSerial(DvcDescs[i].m_Serial);
            if (dr != DTAPI_OK)
                throw Exc(c_ErrFailToAttachToType, Type2Name(DvcDescs[i].m_TypeNumber),
                                                                     DtapiResult2Str(dr));
            int PortNr;
            if (!GetPortWithCaps(&m_DtDvc, ReqCaps, PortNr))
            {
                m_DtDvc.Detach();
                continue;
            }
            FoundDvc = true;
            break;
        }
        if (!FoundDvc)
            throw Exc(c_ErrNoOutputFound);
    }

    // We've found a device with output-ports and are attached to it. Now check if we can
    // find the proper port.
    int  Port = -1;
    
    DtHwFuncDesc HwFuncs[16];
    int  NumHwFuncs;
    dr = m_DtDvc.HwFuncScan(sizeof(HwFuncs)/sizeof(HwFuncs[0]), NumHwFuncs, HwFuncs);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrHwScanFailed, DtapiResult2Str(dr));
    if (m_CmdLineParams.m_Port.IsSet())
    {
        if (m_CmdLineParams.m_Port<1 || m_CmdLineParams.m_Port>NumHwFuncs)
        {
            throw Exc(c_ErrNoSuchPort, Type2Name(m_DtDvc.m_DvcDesc.m_TypeNumber),
                                                          m_CmdLineParams.m_Port.ToInt());
        }
        Port = m_CmdLineParams.m_Port;
        if ((HwFuncs[m_CmdLineParams.m_Port-1].m_Flags&ReqCaps) != ReqCaps)
            throw Exc(c_ErrNoPort, Type2Name(m_DtDvc.m_DvcDesc.m_TypeNumber),
                                                          m_CmdLineParams.m_Port.ToInt());
    } else {
        if (!GetPortWithCaps(&m_DtDvc, ReqCaps, Port))
            throw Exc(c_ErrNoSuitablePort, Type2Name(m_DtDvc.m_DvcDesc.m_TypeNumber));
    }
    // At this point we're sure the chosen port is valid
    dr = m_DtOutp.AttachToPort(&m_DtDvc, Port);
    if (dr != DTAPI_OK)
        throw Exc(c_ErrFailToAttachToChan, ::DtapiResult2Str(dr));
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- A3StlMod::DisplayInfo -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void A3StlMod::DisplayInfo()
{
    char sz[256];

    LogF("");
    LogF("Start ATSC 3.0 STL Modulation");

    if ( m_DtDvc.Category() == DTAPI_CAT_USB )
        sprintf(sz, "DTU-%d", m_DtDvc.TypeNumber());
    else /*if ( m_DtDvc.Category() == DTAPI_CAT_PCI )*/
        sprintf(sz, "DTA-%d", m_DtDvc.TypeNumber());

    if (m_CmdLineParams.m_Port != -1)
        LogF("- Device\t\t: %s port %d (#%d)", sz, m_CmdLineParams.m_Port.ToInt(),
                                                       m_CmdLineParams.m_DvcNum.ToInt());
    else
        LogF("- Device\t\t: %s (#%d)", sz, m_CmdLineParams.m_DvcNum.ToInt());

    LogF("- Carrier Frequency\t: %.2f MHz", m_CmdLineParams.m_CarrierFreq.ToDouble());

    LogF("- Bandwidth\t\t: %d MHz", m_CmdLineParams.m_Bandwidth.ToInt());

    // Show output level (if supported)
    if ( (m_DtOutp.m_HwFuncDesc.m_Flags & DTAPI_CAP_ADJLVL)!=0 )
        LogF("- Output Level\t\t: %.1f dBm", m_CmdLineParams.m_OutpLevel.ToDouble());

    // Log IP parameters
    DtapiIpAddr2Str( sz, sizeof(sz), m_CmdLineParams.m_IpPars.m_Ip);
    if (NwUtils::IsIpAddressEmpty(m_CmdLineParams.m_IpPars.m_SrcFltIp))
        LogF("- Input IP Address\t: %s:%d", sz, m_CmdLineParams.m_IpPars.m_Port);
    else {
        char sz2[256];
        DtapiIpAddr2Str(sz2, sizeof(sz2), m_CmdLineParams.m_IpPars.m_SrcFltIp);
        LogF("- Input IP Address\t: %s@%s:%d", sz2, sz, m_CmdLineParams.m_IpPars.m_Port);
    }
    if (m_CmdLineParams.m_IpItf.IsSet())
    {
        DtapiIpAddr2Str(sz, sizeof(sz), m_CmdLineParams.m_Nic);
        LogF("- Input NIC\t\t: %s", sz);
    }

    Log("");
    Log("Press any key to stop");
    Log("");
}


// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- A3StlMod::DoYourWork -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int A3StlMod::DoYourWork(int argc, char* argv[])
{
    int RetValue=0;
    try
    {
        try
        {
            // Parse command line options
            m_CmdLineParams.ParseCommandLine(argc, argv);
        }
        catch ( Exc e ) {
            Log(e);
            Log("");
            Log("Use -? option for help");
            return -1;
        }
        catch (DtOptException e) {
            Log(e);
            Log("");
            Log("Use -? option for help");
            return -1;
        }
        catch ( ... ) {
            return -1;
        }
        if ( m_CmdLineParams.m_ShowHelp )
        {
            ShowHelp();
            return RetValue;
        }

        //-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Print start message -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
        LogF("A3StlMod v%d.%d.%d (c) 2022 DekTec Digital Video B.V.\n",
                 A3STLMOD_VERSION_MAJOR, A3STLMOD_VERSION_MINOR, A3STLMOD_VERSION_BUGFIX);
     
        // Init ATSC 3.0 output port
        AttachToPort(DTAPI_CAP_TX_ATSC3);
        __int64 Freq = (__int64)(m_CmdLineParams.m_CarrierFreq.ToDouble()*1E6);
        int Level = (int)(m_CmdLineParams.m_OutpLevel.ToDouble()*10);
        int Bandwidth = m_CmdLineParams.m_Bandwidth.ToInt();
        m_StlMod.Init(m_CmdLineParams.m_IpPars, m_CmdLineParams.m_Nic,
                                                       &m_DtOutp, Freq, Bandwidth, Level);

        // Print start info
        DisplayInfo();

        // Show activity indicator?
        bool ActiveInd = m_CmdLineParams.m_ActiveInd.ToBool();

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Receive/Transmit loop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
        m_StlMod.StlMod(ActiveInd);
    }
    catch( Exc e )
    {
        Log(e);
        RetValue = -1;
    }

    // Detach from channel and device
    m_StlMod.Uninit();
    m_DtOutp.Detach(DTAPI_INSTANT_DETACH);

    m_DtDvc.Detach();

    Log("Stopped");

    return RetValue;
}


// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- A3StlMod::ShowHelp -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void A3StlMod::ShowHelp()
{
    LogF("A3StlMod version: %d.%d.%d\n", A3STLMOD_VERSION_MAJOR, 
                                         A3STLMOD_VERSION_MINOR, A3STLMOD_VERSION_BUGFIX);
    Log( "");
    Log( "Usage:");
    Log( "   A3StlMod [-t type] [-n number] [-i port]\n" \
         "              [-mf frequency] [-mb bandwidth] [-ml level_dbm]\n" \
         "              [-ipa ip_address_and_port]\n" \
         "              [-ipsrc ip_source_specific_address]\n" \
         "              [-ipitf ip_network_interface_address]\n" \
         "              [-a] [-?]");
    Log( "");
    Log("Options:");
    DtOptItem::PrintOptionsHelp(g_CmdOptions, g_NumCmdOptions);
    Log("Examples:");
    Log(" A3StlMod -t 2115 -n 1 -i 1 -mf 578 -mb 6 -ml -15.0 -ipa 239.2.2.2:1234");
}

//+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Application entry point +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int argc, char* argv[])
{
    int RetValue(0);
    A3StlMod A3StlMod;
    try
    {   
        RetValue = A3StlMod.DoYourWork(argc, argv);
    }
    catch(...)
    {
        RetValue = -1;
    }
    return RetValue;
}

