// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* A3StlMod.h *#*#*#*#*#*#*#*#*#*#* (C) 2022 DekTec
//
//

#pragma once

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <stdarg.h>
#include "DtOpt.h"
#include "DTAPI.h"

// Additional includes
#include <WinSock2.h>            // Windows socket API
#include <Ws2tcpip.h>            // MULTICAST

// Windows/Microsoft uses with _ => Linux uses without
#ifndef WINBUILD
#define _vsnprintf vsnprintf
#define _vsnwprintf vswprintf
#endif

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- class Exc -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
class Exc
{
public:
    Exc( const char* pFormat, ...  )
    {
        va_list  ArgList;
        va_start(ArgList, pFormat);
        _vsnprintf(m_ErrorMsg, sizeof(m_ErrorMsg)-1, pFormat, ArgList);
        va_end(ArgList);
    };
    virtual ~Exc() {};

    operator const char* () const { return m_ErrorMsg; }

protected:

    // Error message
    char m_ErrorMsg[1024];
};

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- CommandLineParams -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
class CommandLineParams
{
public:
    CommandLineParams() = default;
    virtual ~CommandLineParams()= default;
    void  ParseCommandLine(int argc, char* argv[]);

    //---- Application parameters ----
    DtOpt  m_DvcType;           // -t: Type of device 
    DtOpt  m_DvcNum;            // -n: Device number
    DtOpt  m_Port;              // -i: Port number of channel to use

    DtOpt  m_CarrierFreq;       // -mf: carrier frequency
    DtOpt  m_Bandwidth;         // -mb: ATSC 3.0 bandwidth
    DtOpt  m_OutpLevel;         // -ml: Modulator output level

    DtOpt  m_IpItf;             // -ipitf: Local IP-address of the interface
    DtOpt  m_IpAddr;            // -ipa: IP address of the multicast group.
    DtOpt  m_IpSrc;             // -ipsrc: IP source address (IGMP v3)
    uint8_t m_Nic[4];           // IP address of network interface
    DtIpPars m_IpPars;          // IP parameters.

    DtOpt  m_ActiveInd;         // -a: show activity indicator
    DtOpt  m_ShowHelp;          // -?: show help
};

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- class NwUtils -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
class NwUtils
{  
public:
    static bool  IsMulticastAddressV4(const unsigned char* pIpAddr);
    static bool  IsIpAddressEmpty(const unsigned char* pIpAddress);
    static void  Ip2InAddr(const unsigned char* pIpAddress, in_addr* pInAddr);
    static void  InAddr2Ip(const in_addr* pInAddr, unsigned char* IpAddress);
};

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Log Functions -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
static void Log(const char* pMessage);
static void Log(const wchar_t* pMessage);
static void LogF(const char* pMessage, ... );
static void LogF(const wchar_t* pMessage, ... );

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- class StlModHelper -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
class StlModHelper
{
public:
    StlModHelper() = default;
    void Init(DtIpPars IpPars, uint8_t* NwItf, DtOutpChannel* pA3Outp, 
                                                __int64 RfFreq, int Bandwidth, int Level);
    void Uninit();
    void StlMod(bool ActiveInd);
private:
    void RxStltpInit(DtIpPars IpPars, uint8_t* pNwItf);
    void TxRfInit(DtOutpChannel* pA3Outp, __int64 RfFreq, int Bandwidth, int Level);
    DtOutpChannel* m_pA3Outp {nullptr};
    int m_FifoSize{-1};
    SOCKET m_RxSocket; 
};

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- A3StlMod -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
class A3StlMod
{
public:
    A3StlMod() = default;
    virtual ~A3StlMod() = default;

    //---- Public operations ----
    int DoYourWork(int argc, char* argv[]);

protected:
    //---- Internal operation ----
    void AttachToPort(DtCaps& ReqCaps);
    void DisplayInfo();
    bool GetPortWithCaps(DtDevice* pDvc, DtCaps& ReqCaps, int& PortNr);
    void ShowHelp();   

    CommandLineParams  m_CmdLineParams;
    DtDevice  m_DtDvc;              // Our device
    DtOutpChannel  m_DtOutp;        // Our output channel
    StlModHelper  m_StlMod;         // STLTP to ATSC 3.0 RF
};

